
import numpy as np
from scipy import stats

from .native.pyrisk import python_estimate_downside_risk, python_estimate_upside_risk
from .native.cythonrisk import cython_estimate_downside_risk, cython_estimate_upside_risk
from .constants import dividing_factors_dict


def estimate_downside_risk(timestamps, prices, target_return, unit='year', lowlevellang='C'):
    dividing_factor = dividing_factors_dict[unit]

    ts = np.array(timestamps, dtype='datetime64[s]')
    ts = np.array(ts, dtype=np.float64) / dividing_factor

    if lowlevellang == 'C':
        return cython_estimate_downside_risk(ts, prices, target_return)
    elif lowlevellang == 'P':
        return python_estimate_downside_risk(ts, prices, target_return)
    else:
        raise ValueError(
            'Unknown low-level language: {}. (Should be "P" (Python), or "C" (Cython))'.format(
                lowlevellang))


def estimate_upside_risk(timestamps, prices, target_return, unit='year', lowlevellang='C'):
    dividing_factor = dividing_factors_dict[unit]

    ts = np.array(timestamps, dtype='datetime64[s]')
    ts = np.array(ts, dtype=np.float64) / dividing_factor

    if lowlevellang == 'C':
        return cython_estimate_upside_risk(ts, prices, target_return)
    elif lowlevellang == 'N':
        return python_estimate_upside_risk(ts, prices, target_return)
    else:
        raise ValueError(
            'Unknown low-level language: {}. (Should be "P" (Python), or "C" (Cython).)'.format(
                lowlevellang))


def estimate_beta(timestamps, prices, market_prices, unit='year'):
    dividing_factor = dividing_factors_dict[unit]

    ts = np.array(timestamps, dtype='datetime64[s]')
    ts = np.array(ts, dtype=np.float64) / dividing_factor
    dt = ts[1:] - ts[:-1]

    dlogprices = np.log(prices[1:] / prices[:-1])
    dlogmarketprices = np.log(market_prices[1:] / market_prices[:-1])
    assert len(dt) == len(dlogprices)
    assert len(dt) == len(dlogmarketprices)

    stockyields = dlogprices / dt
    marketyields = dlogmarketprices / dt

    reg = stats.linregress(x=marketyields, y=stockyields)
    return reg.slope




from .native.pymetrics import python_sharpe_ratio, python_mpt_costfunction, python_mpt_entropy_costfunction
from .native.cythonmetrics import cython_sharpe_ratio, cython_mpt_costfunction, cython_mpt_entropy_costfunction


def sharpe_ratio(weights, r, cov, rf, lowlevellang='C'):
    if lowlevellang == 'C':
        return cython_sharpe_ratio(weights, r, cov, rf)
    elif lowlevellang == 'P':
        return python_sharpe_ratio(weights, r, cov, rf)
    else:
        raise ValueError('Unknown low-level language: {}. (Should be "P" (Python), or "C" (Cython).)'.format(lowlevellang))


def mpt_costfunction(weights, r, cov, rf, lamb, V0=10., lowlevellang='C'):
    if lowlevellang == 'C':
        return cython_mpt_costfunction(weights, r, cov, rf, lamb, V0)
    elif lowlevellang == 'P':
        return python_mpt_costfunction(weights, r, cov, rf, lamb, V0=V0)
    else:
        raise ValueError('Unknown low-level language: {}. (Should be "P" (Python), or "C" (Cython).)'.format(lowlevellang))


def mpt_entropy_costfunction(weights, r, cov, rf, lamb0, lamb1, V=10., lowlevellang='C'):
    if lowlevellang == 'C':
        return cython_mpt_entropy_costfunction(weights, r, cov, rf, lamb0, lamb1, V)
    elif lowlevellang == 'P':
        return python_mpt_entropy_costfunction(weights, r, cov, rf, lamb0, lamb1, V=V)
    else:
        raise ValueError('Unknown low-level language: {}. (Should be "P" (Python), or "C" (Cython).)'.format(lowlevellang))





import unittest

import numpy as np
import pandas as pd

from finsim.tech.ma import get_movingaverage_price_data
from finsim.data.preader import get_yahoofinance_data


class TestMovingAverage(unittest.TestCase):
    def test_sp500(self):
        testdf = get_movingaverage_price_data('VOO', '2016-01-01', '2023-02-01', 200)
        assert testdf.iloc[0]['TimeStamp'].date().strftime('%Y-%m-%d') == '2016-01-04'

    def test_abnb(self):
        testdf = get_movingaverage_price_data('ABNB', '2016-01-01', '2023-02-01', 200)
        assert testdf.iloc[0]['TimeStamp'].date().strftime('%Y-%m-%d') == '2021-06-29'

    def test_average(self):
        testmadf = get_movingaverage_price_data('^GSPC', '2016-01-01', '2023-02-01', 200)
        testdf = get_yahoofinance_data('^GSPC', '2016-01-01', '2023-02-01')

        # get last row
        lastrow_stockdata = testdf.iloc[len(testdf)-1].to_dict()
        lastday_considered = lastrow_stockdata['TimeStamp']
        firstday_considered = lastday_considered - pd.Timedelta('200 days')
        ma = np.mean(testdf.loc[
                         (testdf['TimeStamp'] > firstday_considered) & (testdf['TimeStamp'] <= lastday_considered)
                     ]['Close'])

        # comparison
        malastrow_data = testmadf.iloc[len(testmadf)-1].to_dict()
        assert malastrow_data['TimeStamp'] == testdf.iloc[len(testdf)-1, :]['TimeStamp']
        assert malastrow_data['MA'] == ma


if __name__ == '__main__':
    unittest.main()




import logging
from functools import partial
from itertools import product
from datetime import datetime

import numpy as np
from scipy.optimize import minimize
from tqdm import tqdm

from .metrics import sharpe_ratio, mpt_costfunction, mpt_entropy_costfunction
from ..helper import align_timestamps_stock_dataframes
from ...data import get_yahoofinance_data
from ...data.preader import get_dividends_df
from ...estimate.fit import fit_multivariate_BlackScholesMerton_model, fit_BlackScholesMerton_model, \
    fit_timeweighted_BlackScholesMerton_model, fit_timeweighted_multivariate_BlackScholesMerton_model


def getarrayelementminusminvalue(array, minvalue, index):
    return array[index] - minvalue


def checksumarray(array, total):
    return total - np.sum(array)


def optimized_portfolio_on_sharperatio(r, cov, rf, minweight=0.):
    func = partial(sharpe_ratio, r=r, cov=cov, rf=rf)
    nbstocks = len(r)
    initialguess = np.repeat(1 / nbstocks, nbstocks)
    constraints = [
        {'type': 'ineq', 'fun': partial(getarrayelementminusminvalue, minvalue=minweight, index=i)}
        for i in range(nbstocks)
    ] + [
        {'type': 'eq', 'fun': partial(checksumarray, total=1.)}
    ] + [
        {'type': 'ineq', 'fun': lambda weights: weights[i]}
        for i in range(len(r))
    ]
    return minimize(
        lambda weights: -func(weights),
        initialguess,
        constraints=constraints
    )


def optimized_portfolio_mpt_costfunction(r, cov, rf, lamb, V0=10.):
    func = partial(mpt_costfunction, r=r, cov=cov, rf=rf, lamb=lamb, V0=V0)
    nbstocks = len(r)
    constraints = [
        {'type': 'ineq', 'fun': partial(getarrayelementminusminvalue, minvalue=0., index=i)}
        for i in range(nbstocks+1)
    ] + [
        {'type': 'ineq', 'fun': partial(checksumarray, total=V0)}
    ] + [
        {'type': 'ineq', 'fun': lambda weights: weights[i]}
        for i in range(len(r))
    ]
    initialguess = np.repeat(V0 / (nbstocks+1), nbstocks+1)
    return minimize(
        lambda weights: -func(weights),
        initialguess,
        constraints=constraints
    )


def optimized_portfolio_mpt_entropy_costfunction(r, cov, rf, lamb0, lamb1, V=10.):
    func = partial(mpt_entropy_costfunction, r=r, cov=cov, rf=rf, lamb0=lamb0, lamb1=lamb1, V=V)
    nbstocks = len(r)
    constraints = [
        {'type': 'ineq', 'fun': partial(getarrayelementminusminvalue, minvalue=0., index=i)}
        for i in range(nbstocks+1)
    ] + [
        {'type': 'ineq', 'fun': partial(checksumarray, total=V)}
    ] + [
        {'type': 'ineq', 'fun': lambda weights: weights[i]}
        for i in range(len(r))
    ]
    initialguess = np.repeat(V / (nbstocks + 1), nbstocks + 1)
    return minimize(
        lambda weights: -func(weights),
        initialguess,
        constraints=constraints
    )


def intermediate_wrangle_stock_df_without_dividends(stock_df):
    stock_df.loc[:, 'EffVal'] = stock_df['Close'] * 1.
    return stock_df


def intermediate_wrangle_stock_df_with_dividends(stock_df, sym):
    dividends_df = get_dividends_df(sym)
    dividends_df = dividends_df.rename(columns={'date': 'TimeStamp'})
    dividends_df.loc[:, 'Cash'] = np.cumsum(dividends_df['Dividends'].ravel())
    stock_df.loc[:, 'TimeStamp'] = stock_df['TimeStamp'].map(lambda ts: datetime.strftime(ts, '%Y-%m-%d'))
    stock_df = stock_df.merge(dividends_df, how='left').ffill().fillna(0)
    stock_df.loc[:, 'EffVal'] = stock_df['Close'] + stock_df['Cash']
    stock_df.loc[:, 'TimeStamp'] = stock_df['TimeStamp'].map(lambda ts: datetime.strptime(ts, '%Y-%m-%d'))
    return stock_df


def get_BlackScholesMerton_stocks_estimation(
        symbols,
        startdate,
        enddate,
        progressbar=True,
        cacheddir=None,
        include_dividends=False
):
    logging.info('Reading financial data...')
    symreadingprogress = tqdm(symbols) if progressbar else symbols
    stocks_data_dfs = [
        get_yahoofinance_data(sym, startdate, enddate, cacheddir=cacheddir)
        for sym in symreadingprogress
    ]

    if include_dividends:
        for i, symbol in enumerate(symbols):
            stocks_data_dfs[i] = intermediate_wrangle_stock_df_with_dividends(stocks_data_dfs[i], symbol)
    else:
        for i in range(len(symbols)):
            stocks_data_dfs[i] = intermediate_wrangle_stock_df_without_dividends(stocks_data_dfs[i])

    # unify the timestamps columns
    logging.info('Unifying timestamps....')
    stocks_data_dfs = align_timestamps_stock_dataframes(stocks_data_dfs)

    # calculating length
    logging.info('Estimating...')
    max_timearray_ref = 0
    maxlen = max(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)))
    minlen = min(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)) if len(stocks_data_dfs) > 0)   # exclude those stocks that do not exist
    absent_stocks = {sym for sym, df in zip(symbols, stocks_data_dfs) if len(df) == 0}
    logging.debug('maxlen = {}; minlen = {}; absent_stocks: {}'.format(maxlen, minlen, ', '.join(absent_stocks)))

    # same length, directly compare
    if maxlen == minlen:
        return fit_multivariate_BlackScholesMerton_model(
            stocks_data_dfs[max_timearray_ref]['TimeStamp'].ravel(),
            np.array([
                stocks_data_dfs[i]['EffVal'].ravel()
                for i in range(len(stocks_data_dfs))
            ])
        )
    else:    # maxlen != minlen:
        rarray = np.zeros(len(symbols))
        covmat = np.zeros((len(symbols), len(symbols)))

        for i, stock_df in enumerate(stocks_data_dfs):
            r, sigma = fit_BlackScholesMerton_model(stock_df['TimeStamp'].ravel(), stock_df['Close'].ravel())
            rarray[i] = r
            covmat[i, i] = sigma*sigma

        for i, j in product(range(len(symbols)), range(len(symbols))):
            stock_df_i = stocks_data_dfs[i]
            stock_df_j = stocks_data_dfs[j]
            smallerlen = min(len(stock_df_i), len(stock_df_j))
            _, cov = fit_multivariate_BlackScholesMerton_model(
                stock_df_i.loc[(len(stock_df_i)-smallerlen):, 'TimeStamp'].ravel(),
                np.array([
                    stock_df_i.loc[(len(stock_df_i)-smallerlen):, 'Close'].ravel(),
                    stock_df_j.loc[(len(stock_df_j)-smallerlen):, 'Close'].ravel()
                ])
            )
            covmat[i, j] = cov[0, 1]
            covmat[j, i] = cov[1, 0]

        return rarray, covmat


def get_stocks_timeweighted_estimation(
        symbols,
        timeweightdf,
        progressbar=True,
        cacheddir=None,
        include_dividends=False
):
    logging.info('Parsing weights according to date')
    startdate = timeweightdf['TimeStamp'][0]
    if isinstance(startdate, datetime):
        startdate = startdate.strftime('%Y-%m-%d')
    enddate = timeweightdf['TimeStamp'][len(timeweightdf) - 1]
    if isinstance(enddate, datetime):
        enddate = enddate.strftime('%Y-%m-%d')

    logging.info('Reading financial data...')
    symreadingprogress = tqdm(symbols) if progressbar else symbols
    stocks_data_dfs = [
        get_yahoofinance_data(sym, startdate, enddate, cacheddir=cacheddir)
        for sym in symreadingprogress
    ]

    if include_dividends:
        for i, symbol in enumerate(symbols):
            stocks_data_dfs[i] = intermediate_wrangle_stock_df_with_dividends(stocks_data_dfs[i], symbol)
    else:
        for i in range(len(symbols)):
            stocks_data_dfs[i] = intermediate_wrangle_stock_df_without_dividends(stocks_data_dfs[i])

    # unify the timestamps columns
    logging.info('Unifying timestamps....')
    stocks_data_dfs = align_timestamps_stock_dataframes(stocks_data_dfs)

    # calculating length
    logging.info('Estimating...')
    max_timearray_ref = 0
    maxlen = max(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)))
    minlen = min(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)) if
                 len(stocks_data_dfs) > 0)  # exclude those stocks that do not exist
    absent_stocks = {sym for sym, df in zip(symbols, stocks_data_dfs) if len(df) == 0}
    logging.debug('maxlen = {}; minlen = {}; absent_stocks: {}'.format(maxlen, minlen, ', '.join(absent_stocks)))

    # same length, directly compare
    if maxlen == minlen and maxlen == len(timeweightdf):
        return fit_timeweighted_multivariate_BlackScholesMerton_model(
            np.array(stocks_data_dfs[max_timearray_ref]['TimeStamp']),
            np.array([
                np.array(stocks_data_dfs[i]['EffVal'])
                for i in range(len(stocks_data_dfs))
            ]),
            timeweightdf['weight'].ravel()
        )
    else:
        rarray = np.zeros(len(symbols))
        covmat = np.zeros((len(symbols), len(symbols)))

        for i, stock_df in enumerate(stocks_data_dfs):
            w_stock_df = stock_df.merge(timeweightdf, on='TimeStamp', how='left')
            r, sigma = fit_timeweighted_BlackScholesMerton_model(
                w_stock_df['TimeStamp'].ravel(),
                w_stock_df['Close'].ravel(),
                w_stock_df['weight'].ravel()
            )
            rarray[i] = r
            covmat[i, i] = sigma * sigma

        for i, j in product(range(len(symbols)), range(len(symbols))):
            w_stock_df_i = stocks_data_dfs[i].merge(timeweightdf, on='TimeStamp', how='left')
            w_stock_df_j = stocks_data_dfs[j].merge(timeweightdf, on='TimeStamp', how='left')
            smallerlen = min(len(w_stock_df_i), len(w_stock_df_j))
            _, cov = fit_timeweighted_multivariate_BlackScholesMerton_model(
                w_stock_df_i.loc[(len(w_stock_df_i) - smallerlen):, 'TimeStamp'].ravel(),
                np.array([
                    w_stock_df_i.loc[(len(w_stock_df_i) - smallerlen):, 'Close'].ravel(),
                    w_stock_df_j.loc[(len(w_stock_df_j) - smallerlen):, 'Close'].ravel()
                ]),
                w_stock_df_i.loc[(len(w_stock_df_i) - smallerlen):, 'weight'].ravel()
            )
            covmat[i, j] = cov[0, 1]
            covmat[j, i] = cov[1, 0]

        return rarray, covmat




import unittest

import numpy as np

from finsim.estimate import risk


class TestRisk(unittest.TestCase):
    def test_nodownward_movement(self):
        ts = np.linspace(0, 10, 11)
        S = np.linspace(1.0, 1.8, 11)
        self.assertAlmostEqual(
            risk.python_estimate_downside_risk(ts, S, 0.0),
            0.0
        )
        self.assertAlmostEqual(
            risk.cython_estimate_downside_risk(ts, S, 0.0),
            0.0
        )

    def test_somedownward_movement(self):
        r_fake_array = np.array([0.1, 0.2, -0.1, -0.1, 0.3, 0.4])
        S_array = np.exp(np.cumsum(r_fake_array))
        ts = np.linspace(0, 5, 6)
        expected_downside_risk = np.sqrt(0.02/5)

        self.assertAlmostEqual(
            risk.python_estimate_downside_risk(ts, S_array, 0.0),
            expected_downside_risk
        )
        self.assertAlmostEqual(
            risk.cython_estimate_downside_risk(ts, S_array, 0.0),
            expected_downside_risk
        )

    def test_noupward_movement(self):
        ts = np.linspace(0, 10, 11)
        S = np.linspace(1.0, 0.5, 11)

        self.assertAlmostEqual(
            risk.python_estimate_upside_risk(ts, S, 0.0),
            0.0
        )
        self.assertAlmostEqual(
            risk.cython_estimate_upside_risk(ts, S, 0.0),
            0.0
        )

    def test_someupward_movement(self):
        r_fake_array = np.array([-0.1, -0.2, 0.1, 0.1, -0.3, -0.4])
        S_array = np.exp(np.cumsum(r_fake_array))
        ts = np.linspace(0, 5, 6)
        expected_upside_risk = np.sqrt(0.02/5)

        self.assertAlmostEqual(
            risk.python_estimate_upside_risk(ts, S_array, 0.0),
            expected_upside_risk
        )
        self.assertAlmostEqual(
            risk.cython_estimate_upside_risk(ts, S_array, 0.0),
            expected_upside_risk
        )

    def test_beta(self):
        timestamps = np.array(['2021-02-16T00:00:00.000000000', 
                               '2021-02-17T00:00:00.000000000', 
                               '2021-02-18T00:00:00.000000000', 
                               '2021-02-19T00:00:00.000000000'], 
                              dtype='datetime64[ns]')
        index_prices = 100*np.exp([0., 0.02, 0.02+0.01, 0.02+0.01-0.015])
        stock_prices = 10*np.exp([0., 0.04, 0.04+0.02, 0.04+0.02-0.03])
        beta = risk.estimate_beta(timestamps, stock_prices, index_prices)
        self.assertAlmostEqual(beta, 2.0)


if __name__ == '__main__':
    unittest.main()




import numpy as np
from itertools import product

from .constants import dividing_factors_dict
from .native.pyfit import python_fit_BlackScholesMerton_model, python_fit_multivariate_BlackScholesMerton_model
from .native.cythonfit import cython_fit_BlackScholesMerton_model, cython_fit_multivariate_BlackScholesMerton_model


# Note: always round-off to seconds first, but flexible about the unit to be used.

def fit_BlackScholesMerton_model(timestamps, prices, unit='year', lowlevellang='C'):
    dividing_factor = dividing_factors_dict[unit]

    ts = np.array(timestamps, dtype='datetime64[s]')
    ts = np.array(ts, dtype=np.float64) / dividing_factor

    if lowlevellang == 'C':
        return cython_fit_BlackScholesMerton_model(ts, prices)
    elif lowlevellang == 'P':
        return python_fit_BlackScholesMerton_model(ts, prices)
    else:
        raise ValueError(
            'Unknown low-level language: {}. (Should be "P" (Python), or "C" (Cython).)'.format(
                lowlevellang))


def fit_multivariate_BlackScholesMerton_model(timestamps, multiprices, unit='year', lowlevellang='C'):
    dividing_factor = dividing_factors_dict[unit]

    ts = np.array(timestamps, dtype='datetime64[s]')
    ts = np.array(ts, dtype=np.float64) / dividing_factor

    if lowlevellang == 'C':
        return cython_fit_multivariate_BlackScholesMerton_model(ts, multiprices)
    elif lowlevellang == 'P':
        return python_fit_multivariate_BlackScholesMerton_model(ts, multiprices)
    else:
        raise ValueError(
            'Unknown low-level language: {}. (Should be "C" (Cython).)'.format(
                lowlevellang))


######## routines below are for time-weighted portfolio building

def fit_timeweighted_BlackScholesMerton_model(timestamps, prices, weights, unit='year'):
    dividing_factor = dividing_factors_dict[unit]

    ts = np.array(timestamps, dtype='datetime64[s]')
    ts = np.array(ts, dtype=np.float64) / dividing_factor

    dlogS = np.log(prices[1:] / prices[:-1])
    dt = ts[1:] - ts[:-1]

    r = np.average(dlogS / dt, weights=weights[1:])
    sigma = np.sqrt(np.average(np.square(dlogS / np.sqrt(dt)), weights=weights[1:]) - np.square(
        np.average(dlogS / np.sqrt(dt), weights=weights[1:])))

    return r, sigma


def fit_timeweighted_multivariate_BlackScholesMerton_model(timestamps, multiprices, weights, unit='year'):
    dividing_factor = dividing_factors_dict[unit]

    ts = np.array(timestamps, dtype='datetime64[s]')
    ts = np.array(ts, dtype=np.float64) / dividing_factor

    # possibly hit this part
    dlogS = np.log(multiprices[:, 1:] / multiprices[:, :-1])
    dt = ts[1:] - ts[:-1]

    # estimation
    r = np.zeros(multiprices.shape[0])
    for i in range(multiprices.shape[0]):
        r[i] = np.average((dlogS[i, :] / dt), weights=weights[1:])
    cov = np.zeros((multiprices.shape[0], multiprices.shape[0]))
    for i, j in product(range(multiprices.shape[0]), range(multiprices.shape[0])):
        avg_i = np.average(dlogS[i, :] / np.sqrt(dt), weights=weights[1:])
        avg_j = np.average(dlogS[j, :] / np.sqrt(dt), weights=weights[1:])
        cov[i, j] = np.average(dlogS[i, :] * dlogS[j, :] / dt, weights=np.square(weights[1:])) - avg_i * avg_j

    return r, cov




import numpy as np


def python_sharpe_ratio(weights, r, cov, rf):
    yieldrate = np.sum(weights * r)
    sqweights = np.expand_dims(weights, axis=1) @ np.expand_dims(weights, axis=0)
    volatility = np.sqrt(np.sum(sqweights * cov))
    return (yieldrate - rf) / volatility


def python_mpt_costfunction(weights, r, cov, rf, lamb, V0=10.):
    weightmat = np.expand_dims(weights[:-1], axis=0)
    c = lamb * V0
    return weights[-1]*rf + np.dot(weights[:-1], r) - 0.5*c/V0*(weightmat @ cov @ weightmat.T)[0, 0]


def python_mpt_entropy_costfunction(weights, r, cov, rf, lamb0, lamb1, V=10.):
    weightmat = np.expand_dims(weights[:-1], axis=0)
    c0 = lamb0 * V
    c1 = lamb1 * V
    yield_val = weights[-1]*rf + np.dot(weights[:-1], r)
    cov_val = - 0.5 * c0 / V * (weightmat @ cov @ weightmat.T)[0, 0]
    sumweights = np.sum(weights[:-1])
    entropy_val = - 0.5 * c1 / V * np.sum(weights[:-1] * (np.log(weights[:-1]) - np.log(sumweights))) / sumweights
    return yield_val + cov_val + entropy_val




from datetime import datetime, timedelta
import os
import logging
from time import sleep
import glob
from functools import lru_cache

import pandas as pd
import yfinance as yf
import tables as tb
from tqdm import tqdm


def extract_online_yahoofinance_data(symbol, startdate, enddate):
    try:
        df = yf.download(
            symbol,
            start=datetime.strptime(startdate, '%Y-%m-%d'),
            end=datetime.strptime(enddate, '%Y-%m-%d') + timedelta(days=1)
        )
    except:
        logging.warning('Symbol {} does not exist between {} and {}.'.format(symbol, startdate, enddate))
        return pd.DataFrame({
            'TimeStamp': [],
            'High': [],
            'Low': [],
            'Open': [],
            'Close': [],
            'Adj Close': [],
            'Volume': [],
        })

    oricols = df.columns
    df['TimeStamp'] = df.index
    # df['TimeStamp'] = df['TimeStamp'].dt.strftime('%Y-%m-%d')
    # df['Date'] = df['TimeStamp'].apply(lambda ts: ts.date())
    df = df[['TimeStamp'] + list(oricols)]

    return df


def extract_batch_online_yahoofinance_data(symbols, startdate, enddate, threads=True):
    combined_df = yf.download(
        ' '.join(symbols),
        start=datetime.strptime(startdate, '%Y-%m-%d'),
        end=datetime.strptime(enddate, '%Y-%m-%d') + timedelta(days=1),
        group_by='ticker',
        threads=threads
    )

    dataframes = {}
    for symbol in symbols:
        try:
            df = combined_df[symbol].copy()
            oricols = df.columns
            df['TimeStamp'] = df.index
            # df['TimeStamp'] = df['TimeStamp'].dt.strftime('%Y-%m-%d')
            df = df[['TimeStamp'] + list(oricols)]
            dataframes[symbol] = df
        except:
            dataframes[symbol] = pd.DataFrame({
                'TimeStamp': [],
                'High': [],
                'Low': [],
                'Open': [],
                'Close': [],
                'Adj Close': [],
                'Volume': [],
            })

    return dataframes


# yahoo reader with local cache
METATABLE_FILENAME = 'SYMBOL_CACHE.h5'
METATABLE_ROWDES = {
    'symbol': tb.StringCol(shape=(), itemsize=10, dflt='', pos=0),
    'query_startdate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=1),
    'query_enddate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=2),
    'data_startdate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=3),
    'data_enddate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=4)
}


def get_yahoofinance_data(symbol, startdate, enddate, cacheddir=None):
    if cacheddir is None:
        return extract_online_yahoofinance_data(symbol, startdate, enddate)

    if isinstance(cacheddir, str):
        if not os.path.exists(cacheddir) or (os.path.exists(cacheddir) and not os.path.isdir(cacheddir)):
            logging.info('Creating directory: {}'.format(cacheddir))
            os.makedirs(cacheddir)
        if not os.path.exists(os.path.join(cacheddir, METATABLE_FILENAME)):
            logging.info('Creating file: {}'.format(os.path.join(cacheddir, METATABLE_FILENAME)))
            metatable_h5file = tb.open_file(os.path.join(cacheddir, METATABLE_FILENAME), 'w')
            table = metatable_h5file.create_table('/', 'metatable', METATABLE_ROWDES, title='metatable')
        else:
            metatable_h5file = tb.open_file(os.path.join(cacheddir, METATABLE_FILENAME), 'r+')
            table = metatable_h5file.root.metatable

        preexist = False
        for row in table.where('symbol=="{}"'.format(symbol)):
            preexist = True
            # print("{} <= {}: {}".format(row['query_startdate'].decode('utf-8'), startdate,
            #                                     row['query_startdate'].decode('utf-8') <= startdate))
            # print("{} <= {}: {}".format(row['query_enddate'].decode('utf-8'), enddate,
            #                                     row['query_enddate'].decode('utf-8') >= enddate))
            if row['query_startdate'].decode('utf-8') <= startdate and row['query_enddate'].decode('utf-8') >= enddate:
                df = pd.read_hdf(os.path.join(cacheddir, '{}.h5'.format(symbol)), 'yahoodata')
                if len(df) > 0:
                    df = df[(df['TimeStamp'] >= startdate) & (df['TimeStamp'] <= enddate)]
                metatable_h5file.close()
                df = df[~df['Close'].isna()]
                return df

        df = extract_online_yahoofinance_data(symbol, startdate, enddate)
        logging.debug('Caching data for {} from {} to {}'.format(symbol, startdate, enddate))
        df.to_hdf(os.path.join(cacheddir, '{}.h5'.format(symbol)), 'yahoodata')

        if preexist:
            logging.debug('Updating symbol {} in metatable.'.format(symbol))
            for row in table.where('symbol=="{}"'.format(symbol)):
                row['query_startdate'] = startdate
                row['query_enddate'] = enddate
                if len(df) > 0:
                    row['data_startdate'] = datetime.strftime(df['TimeStamp'][0].date(), '%Y-%m-%d')
                    row['data_enddate'] = datetime.strftime(df['TimeStamp'][-1].date(), '%Y-%m-%d')
                else:
                    row['data_startdate'] = '0000-00-00'
                    row['data_enddate'] = ' 0000-00-00'
                row.update()
        else:
            logging.debug('Creating symbol {} in metatable'.format(symbol))
            newrow = table.row
            newrow['symbol'] = symbol
            newrow['query_startdate'] = startdate
            newrow['query_enddate'] = enddate
            if len(df) > 0:
                newrow['data_startdate'] = datetime.strftime(df['TimeStamp'][0].date(), '%Y-%m-%d')
                newrow['data_enddate'] = datetime.strftime(df['TimeStamp'][-1].date(), '%Y-%m-%d')
            else:
                newrow['data_startdate'] = '0000-00-00'
                newrow['data_enddate'] = '0000-00-00'
            newrow.append()

        table.flush()
        metatable_h5file.close()

        df = df[~df['Close'].isna()]

        return df
    else:
        raise TypeError('Type of cacheddir has to be str, but got {} instead!'.format(type(cacheddir)))


@lru_cache(maxsize=256)
def get_symbol_closing_price(symbol, datestr, epsilon=1e-10, cacheddir=None, backtrack=False):
    try:
        return get_yahoofinance_data(symbol, datestr, datestr, cacheddir=cacheddir)['Close'][0]
    except IndexError as e:
        if backtrack:
            prevdatestr = datetime.strftime(datetime.strptime(datestr, '%Y-%m-%d') - timedelta(days=1), '%Y-%m-%d')
            return get_symbol_closing_price(symbol, prevdatestr, epsilon=epsilon, cacheddir=cacheddir, backtrack=True)
        else:
            raise e
    except KeyError:
        return epsilon


def finding_missing_symbols_in_cache(symbols, startdate, enddate, cacheddir):
    if not os.path.exists(os.path.join(cacheddir, METATABLE_FILENAME)):
        return symbols

    # in table
    metatable = pd.read_hdf(os.path.join(cacheddir, METATABLE_FILENAME), 'metatable')
    existing_within_range_symbols = list(
        metatable['symbol'][
            (metatable['query_startdate'] <= startdate) & (metatable['query_enddate'] >= enddate)
        ]
    )
    if logging.root.level >= logging.DEBUG:
        logging.debug('exisiting within range symbols')
        for symbol in existing_within_range_symbols:
            logging.debug('\t{}'.format(symbol))

    # check what are in the cached directories
    existing_symbols = [
        os.path.basename(filepath)
        for filepath in glob.glob(os.path.join(cacheddir, '*.h5'))
    ]
    existing_symbols = [filename[:-3] for filename in existing_symbols if filename != METATABLE_FILENAME]
    if logging.root.level >= logging.DEBUG:
        logging.debug('exisiting symbols')
        for symbol in existing_symbols:
            logging.debug('\t{}'.format(symbol))

    existing_valid_symbols = set(existing_within_range_symbols) & set(existing_symbols)

    return sorted(list(set(symbols) - set(existing_valid_symbols)))


def generating_cached_yahoofinance_data(symbols, startdate, enddate, cacheddir, slicebatch=50, waittime=1, threads=True):
    tocache_symbols = finding_missing_symbols_in_cache(symbols, startdate, enddate, cacheddir)

    logging.info('Total number of symbols: {}'.format(len(symbols)))
    logging.info('Total number of symbols needed to cache: {}'.format(len(tocache_symbols)))
    if not os.path.exists(cacheddir) or (os.path.exists(cacheddir) and not os.path.isdir(cacheddir)):
        logging.info('Creating directory: {}'.format(cacheddir))
        os.makedirs(cacheddir)
    if not os.path.exists(os.path.join(cacheddir, METATABLE_FILENAME)):
        logging.info('Creating file: {}'.format(os.path.join(cacheddir, METATABLE_FILENAME)))
        metatable_h5file = tb.open_file(os.path.join(cacheddir, METATABLE_FILENAME), 'w')
        table = metatable_h5file.create_table('/', 'metatable', METATABLE_ROWDES, title='metatable')
    else:
        metatable_h5file = tb.open_file(os.path.join(cacheddir, METATABLE_FILENAME), 'r+')
        table = metatable_h5file.root.metatable

    nbsymbols = len(tocache_symbols)
    for startidx in tqdm(range(0, nbsymbols, slicebatch)):
        success = False
        while not success:
            try:
                dataframes = extract_batch_online_yahoofinance_data(
                    tocache_symbols[startidx:min(startidx + slicebatch, nbsymbols)],
                    startdate,
                    enddate,
                    threads=threads
                )
                success = True
            except:
                sleep(waittime)

        for symbol in dataframes:
            df = dataframes[symbol]
            df = df[~df['Close'].isna()]
            logging.debug('Caching data for {} from {} to {}'.format(symbol, startdate, enddate))
            df.to_hdf(os.path.join(cacheddir, '{}.h5'.format(symbol)), 'yahoodata')

            logging.debug('Creating symbol {} in metatable'.format(symbol))
            newrow = table.row
            newrow['symbol'] = symbol
            newrow['query_startdate'] = startdate
            newrow['query_enddate'] = enddate
            if len(df) > 0:
                newrow['data_startdate'] = datetime.strftime(df['TimeStamp'][0].date(), '%Y-%m-%d')
                newrow['data_enddate'] = datetime.strftime(df['TimeStamp'][-1].date(), '%Y-%m-%d')
            else:
                newrow['data_startdate'] = '0000-00-00'
                newrow['data_enddate'] = '0000-00-00'
            newrow.append()

        table.flush()

    metatable_h5file.close()


@lru_cache(maxsize=30)
def get_dividends_df(symbol):
    ticker = yf.Ticker(symbol)
    df = pd.DataFrame(ticker.dividends)
    df['TimeStamp'] = df.index.map(lambda item: datetime.strftime(item, '%Y-%m-%d'))
    df = df[['TimeStamp', 'Dividends']]
    return df
